      program driver
c=====7==1=========2=========3=========4=========5=========6=========7==
c
c     SPMD-style driver for Parallel Iterativ Methods (PIM) programs
c                   
      INCLUDE "include_parameters"

c=============declarations for vectors==================================
      real                rhs(veclen), x(veclen), xref(veclen), 
     :                    work(6*veclen)
      common            / vectors / rhs, x, xref, work

c=============declarations for parameter arrays=========================
      integer             iparsize, sparsize, matparsize 
      parameter         ( iparsize = 13, sparsize = 6 , matparsize = 30)
      integer             ipar(iparsize), matpar(matparsize), diag(10)
      real                spar(sparsize)

c=============declarations for parameter array values===================
      integer             nmat, nloc, nproc, mypid,
     :                    istop, iter_max, niter, istat, istep, 
     :                    nzr, nspread, firstrow
      real                tol, tol_f

c=============declarations for diagnostic and timing====================
      integer             nzero_min, nzero_max, nzero_sum,
     :                    nspread_min, nspread_max, nspread_sum,
     :                    nfill_min, nfill_max, 
     :                    ncom_min, ncom_max, ncom_sum,
     :                    nloc_min, nloc_max
      real                nloc_mean, nzero_mean, nfill_mean, 
     :                    nspread_mean, ncom_mean      
      integer             noper_pim, noper_matvec
      real                t_pim_min, t_pim_max, speed_pim,
     :                    t_matvec_min, t_matvec_max, speed_matvec,
     :                    t_ini_min, t_ini_max,
     :                    ets, lib_timer

c=============declarations for other variables====================
      integer             jp, i, nstride, nend, ierr

c=============declarations for external functions and subroutines=======
      real                psnrm
      external            matvec, preconl, preconr, pssum, psnrm, 
     :                    progress
c==============end of declarations======================================

c============= initialize communication context ======================== 
      call lib_initialize(nproc,mypid,ierr)

c============== input and parameter check ==============================
      if (mypid.eq.0) then
         read(5,*) nmat
         read(5,*) nzr
         read(5,*) nspread
         read(5,*) iter_max
         read(5,*) tol
         read(5,*) istop 
      end if

      call lib_brdcst_i(nmat,1,0,ierr)
      call lib_brdcst_i(nzr,1,0,ierr)
      call lib_brdcst_i(nspread,1,0,ierr)
      call lib_brdcst_i(iter_max,1,0,ierr)
      call lib_brdcst_s(tol,1,0,ierr)
      call lib_brdcst_i(istop,1,0,ierr)
      call lib_barrier()


c---------- setup and distribute matrix, rhs ans startvector
      matpar(2) = nmat
      matpar(3) = nzr
      matpar(4) = nspread
      matpar(11) = nproc
      matpar(12) = mypid
      call setupmatrix(matpar,rhs,x)

      if (matpar(21).lt.0) then
         call lib_barrier()
         write(6,*)mypid,' status in matrix generation:',matpar(21)
         call lib_barrier()
         go to 80
      end if
      nmat = matpar(2)
      firstrow = matpar(15)
      nloc = matpar(13)

c---------- initialize parameters ipar and dpar 
c---------- ipar(2) = nmat  (number of rows or columns of matrix)
c---------- ipar(4) = nloc  (number of local rows) 
c---------- ipar(5) = nrst  (restarting parameter)
c---------- ipar(6) = nproc (number of processors)
c---------- ipar(7) = mypid (processor id)
c---------- ipar(8) = iprec (information for preconditioning)
c---------- ipar(9) = istop (stopping criterion)
c---------- ipar(10) = iter_max (maximal number of iterations)
c---------- spar(1) = tol (value for stopping criterion)

      call pimssetpar(ipar,spar,
     :                0,nmat,0,nloc,1,nproc,mypid,
     :                0,istop,iter_max,tol)

c---------- initialize communication structure for matvec
      call lib_barrier()
      ets = lib_timer()

      call matvec_ini(ipar,diag)

      ets = lib_timer() - ets
      call lib_barrier()
      call lib_globmax_s(ets,t_ini_max,1,0,ierr)
      call lib_globmin_s(ets,t_ini_min,1,0,ierr)

c---------- generate diagnostics for matrix structure
      call lib_globsum_i(diag(2),nzero_sum,1,0,ierr)
      call lib_globmin_i(diag(2),nzero_min,1,0,ierr)
      call lib_globmax_i(diag(2),nzero_max,1,0,ierr)
      call lib_globmin_i(diag(3),nfill_min,1,0,ierr)
      call lib_globmax_i(diag(4),nfill_max,1,0,ierr)
      call lib_globsum_i(diag(5),nspread_sum,1,0,ierr)
      call lib_globmin_i(diag(6),nspread_min,1,0,ierr)
      call lib_globmax_i(diag(7),nspread_max,1,0,ierr)
      call lib_globsum_i(diag(8),ncom_sum,1,0,ierr)
      call lib_globmin_i(diag(8),ncom_min,1,0,ierr)
      call lib_globmax_i(diag(8),ncom_max,1,0,ierr)
      call lib_globmin_i(nloc,nloc_min,1,0,ierr)
      call lib_globmax_i(nloc,nloc_max,1,0,ierr)
     
c-------------- set up right hand side for solution x(i) = i
      do i = 1 , nloc
         x(i) = firstrow - 1 + i
      end do
      call matvec(x,rhs,ipar)

c-------------- timing matvec
      call lib_barrier()
      ets = lib_timer()
      call matvec(x,rhs,ipar)
      ets = lib_timer() - ets
      call lib_barrier()
      call lib_globmax_s(ets,t_matvec_max,1,0,ierr)
      call lib_globmin_s(ets,t_matvec_min,1,0,ierr)

c------------- check matvec results
      call matvec_check(rhs,x,xref,ipar)
      call lib_barrier()

c------------- diogonal scaling of matrix and right hand side
      call generate_scale(ipar)
      call scale_matrix(ipar)
      call scale_vector (rhs,ipar)

c------------- initialize start vector x(i) = 0
      do i = 1 , nloc
         x(i) = 0
      end do
c------------- execute and time parallel algorithm
      call lib_barrier()
      ets = lib_timer()
c      call pimsrgmres(x,rhs,work,ipar,spar,
c     :                matvec,preconl,preconr,pssum,psnrm,progress)
      call pimscg(x,rhs,work,ipar,spar,
     :            matvec,preconl,preconr,pssum,psnrm,progress)

      ets = lib_timer() - ets
      call lib_barrier()
      call lib_globmax_s(ets,t_pim_max,1,0,ierr)
      call lib_globmin_s(ets,t_pim_min,1,0,ierr)

c------------- compare computed solution to exact result
      call lib_barrier()
c------------- diogonal scaling of solution
      call scale_vector (x,ipar)
      do jp = 0 , nproc-1
         call lib_barrier()
         if (mypid.eq.jp) then
            firstrow = matpar(15)
            nloc = matpar(13)
            nstride = max(nloc/3,1)
            nend = max(firstrow+nloc-3,firstrow+nstride)
            write(6,*)nloc,'solutions on proc ',mypid
            write(6,1002)(i,i=firstrow,nend,nstride)
            write(6,1003)(x(i),i=1,nend-firstrow+1,nstride)
         end if
         call lib_barrier()
      end do
 1002    format(2x,'i    ',5(i8,14x))
 1003    format(2x,'x(i) ',5f22.14)

      if (mypid.eq.0) then
c============== begin output ===========================================
         niter = ipar(11)
         istat = ipar(12)
         istep = ipar(13)
         tol_f = spar(2)
c--------------Operation count:
         noper_matvec = 2*nzero_sum
         speed_matvec = 0.0
         if (t_matvec_max.gt.0.0) 
     :       speed_matvec = 1.e-6*noper_matvec / t_matvec_max
         noper_pim = niter*( 2*nzero_sum + 2*nmat*(6) )
         speed_pim = 0.0
         if (t_pim_max.gt.0.0) 
     :       speed_pim = 1.e-6*noper_pim / t_pim_max

         nloc_mean = real(nmat) / real(nproc)
         nzero_mean = real(nzero_sum) / real(nproc)
         nfill_mean = real(nzero_sum) / real(nmat)
         nspread_mean = real(nspread_sum) / real(nmat)
         ncom_mean = real(ncom_sum) / real(nproc)
         write(6,*)
         write(6,*)'=========== output of information ================='
         write(6,'(A40,i10)')
     :        'number of processors                 : ',
     :         nproc 
         write(6,*)
         write(6,*)
     :        'matrix properties:--------------------------------------'
         write(6,'(A40,i12)')
     :        'global dimension                     : ',
     :         nmat
         write(6,'(A40,i12,f12.1,i12)')
     :        'local rows : min, mean, max          : ',
     :         nloc_min, nloc_mean, nloc_max
         write(6,'(A40,i12)')
     :        'number of nonzeros                   : ',
     :         nzero_sum
         write(6,'(A40,i12,f12.1,i12)')
     :        'local nonzeros: min, mean, max       : ',
     :         nzero_min, nzero_mean, nzero_max
         write(6,'(A40,i12,f12.1,i12)')
     :        'rowfill : min, mean, max             : ',
     :         nfill_min, nfill_mean, nfill_max
         write(6,'(A40,i12,f12.1,i12)')
     :        'rowspread : min, mean, max           : ',
     :         nspread_min, nspread_mean, nspread_max
         write(6,'(A40,i10)')
     :        'total communication                  : ',
     :         ncom_sum
         write(6,'(A40,i12,f12.1,i12)')
     :        'local comm.: min, mean, max          : ',
     :         ncom_min, ncom_mean, ncom_max
         write(6,'(A40,i12)')
     :        'communication steps                  : ', 
     :         diag(9)
         write(6,*)
         write(6,*)
     :        'solution properties:------------------------------------'
         write(6,'(A40,i12)')
     :        'stopping criterion                   : ',istop
         write(6,'(A40,2i12)')
     :        'iterations: maximal, final           : ',iter_max, niter
         write(6,'(A40,2i12)')
     :        'exit status, breakdown in step       : ',istat,istep
         write(6,'(A40,2e12.3)')
     :        'required tolerance: required, final  : ',tol, tol_f
         write(6,*)
         write(6,*)
     :        'performance properties:---------------------------------'
         write(6,'(A40,2e12.3)')
     :        'time for pim, spread [secs]          : ',
     :         t_pim_max,t_pim_max-t_pim_min
         write(6,'(A40,2f12.2)')
     :        'speed [Mflop/s]: total, per-proc.    : ',
     :         speed_pim, speed_pim/nproc
         write(6,'(A40,2e12.3)')
     :        'time for matvec, spread [secs]       : ',
     :         t_matvec_max,t_matvec_max-t_matvec_min
         write(6,'(A40,2f12.2)')
     :        'speed [Mflop/s]: total, per-proc.    : ',
     :         speed_matvec, speed_matvec/nproc
         write(6,'(A40,2e12.3)')
     :        'time for matvec_ini, spread [secs]   : ',
     :         t_ini_max,t_ini_max-t_ini_min
      end if
   80 continue

      call lib_finalize(nproc,mypid,ierr)

      stop
      end


      subroutine output(matpar,rhs,x)
c=====7==1=========2=========3=========4=========5=========6=========7==
      implicit none
      integer       nproc, mypid
      real      rhs(*), x(*)
      integer   matpar(*)

      INCLUDE "include_parameters"
c------------- data structure for distributed matrix in 
c------------- global address space 
      integer             col(matlen), rpt(veclen+1)
      real                val(matlen)
      common            / matrix / val, col, rpt
      integer nloc, nzloc, ip, i

      nproc = matpar(11)
      mypid = matpar(12)
      nloc = matpar(13)
      nzloc = matpar(14)
      do ip = 0 , nproc - 1
         call lib_barrier()
         if (mypid.eq.ip) then
            write(6,*) 'on proc', ip, nloc, nzloc
            write(6,*) 'val---------'
            write(6,101) (val(i), i = 1,nzloc)
            write(6,*) 'col---------'
            write(6,102) (col(i), i = 1,nzloc)
            write(6,*) 'rpt---------'
            write(6,102) (rpt(i), i = 1,nloc+1)
 1          write(6,*) 'rhs---------'
            write(6,*) (rhs(i), i = 1,nloc)
         end if
         call lib_barrier()
      end do
 101  format(2x,20f4.1)
 102  format(2x,20i4)

      return
      end









