      subroutine setupmatrix (matpar,rhs,x)
      implicit none
 

      integer matpar(*)
      double precision rhs(*), x(*)
c=============declarations for problem definition=======================
c-------------parameters------------------------------------------------

      integer             nmat_max
      parameter         ( nmat_max = 1 00 000 )
      integer             nzero_max
      parameter         ( nzero_max = 1 000 000 )
      integer             col_s(nzero_max), rpt_s(nmat_max+1)
      double precision    val_s(nzero_max) 
      double precision    rhs_s(nmat_max), x_s(nmat_max)

      INCLUDE "include_parameters"
c------------- data structure for distributed matrix in 
c------------- global address space 
      integer             col(matlen), rpt(veclen+1)
      real                val(matlen)
      common            / matrix / val, col, rpt

      integer             nzero, ngrid
      integer mattype,nmat,nz_row,nspread,nproc, ierr

c------------- local variables

      integer firstrow_on_proc(proclen+1)
      integer firstrow, firstval, nloc, nloc1, nzloc
      integer firstrowp, firstvalp, nlocp, nzlocp
      integer i, ip, mypid

      matpar(21) = 0
      mattype = matpar(1)
      nmat = matpar(2)
      nz_row = matpar(3)
      nspread = matpar(4)
      nproc = matpar(11)
      mypid = matpar(12)

      if (nproc.gt.proclen) then 
         if (mypid.eq.1) then
            write(6,*) 'nproc =',nproc,' is too large'
            write(6,*) 'maximal value is proclen =',proclen
         end if
         call lib_barrier()
         matpar(21) = -1
         return
      end if

      call lib_barrier()

      if (mypid.eq.0) then

         ngrid = 1 + sqrt(real(nmat+1))
         nmat = nmat_max
         nzero = nzero_max
         call problem_fd2d(ngrid,nmat,nzero,
     :                        val_s,col_s,rpt_s,rhs_s,matpar(21))

         do i = 1 , nmat
            x_s(i) = 0.0
         end do  

         call distribute_index(nmat,nproc,firstrow_on_proc)
         write(6,*)'distribution',(firstrow_on_proc(i),i=1,nproc+1)
      end if

      call lib_barrier()

      call lib_brdcst_i(nmat,1,0,ierr)
      call lib_brdcst_i(firstrow_on_proc,nproc+1,0,ierr)

      do ip = 0 , nproc-1
         if (mypid.eq.0) then
            firstrowp = firstrow_on_proc(ip+1)
            nlocp = firstrow_on_proc(ip+2) - firstrow_on_proc(ip+1) 
            firstvalp = rpt_s(firstrowp)
            nzlocp = rpt_s(firstrowp+nlocp) - rpt_s(firstrowp)
         end if
         call lib_p2p_s(val_s(firstvalp),nzlocp,0,
     :                       val,nzloc,ip,ierr)
         call lib_p2p_i(col_s(firstvalp),nzlocp,0,
     :                       col,nzloc,ip,ierr)
         call lib_p2p_i(rpt_s(firstrowp),nlocp+1,0,
     :                       rpt,nloc1,ip,ierr)
         call lib_p2p_s(rhs_s(firstrowp),nlocp,0,
     :                       rhs,nloc,ip,ierr)
         call lib_p2p_s(x_s(firstrowp),nlocp,0,
     :                       x,nloc,ip,ierr)
      end do

c----------- let rpt point to the _local_ row indices  
c----------- of the individual processors
      firstval = rpt(1)
      do i = 1, nloc+1
         rpt(i) = rpt(i) - firstval+1
      end do

      matpar(2) = nmat
      matpar(13) = nloc
      matpar(14) = nzloc
      matpar(15) = firstrow_on_proc(mypid+1)

      return 
      end 


      subroutine distribute_index(nmat,nproc,firstrow_on_proc)
      implicit none
      integer nmat, nproc, firstrow_on_proc(*)

      integer nloc, nproc1, ip
      firstrow_on_proc(1) = 1

      nloc = (nmat+nproc-1) / nproc
      nproc1 = nloc*nproc - nmat
      do ip = 1 , nproc-nproc1
         firstrow_on_proc(ip+1) = firstrow_on_proc(ip) + nloc
      end do
      do ip = nproc-nproc1+1 , nproc
         firstrow_on_proc(ip+1) = firstrow_on_proc(ip) + nloc-1
      end do

      return
      end





      subroutine problem_fd2d(n,nmat,nzero,val,col,rpt,rhs,istat)
c=====7==1=========2=========3=========4=========5=========6=========7==
c  Das Unterprogramm zimmer diskretisiert die zweidimensionale 
c  stationaere Waermeleitungsgleichung
c  
c        0 = kappa * laplace(u) + 1/(c*rho) * q
c
c  wobei kappa die Temperaturleitzahl lambda/(c*rho), 
c  lambda die Waermeleitzahl, c die spezifische Waerme, rho die Dichte,
c  u die Temperatur und q die Dichte der stationaeren Waermequelle ist.
c  Das betrachtete Gebiet ist ein zweidimensionaler Schnitt durch ein
c  kubisches Zimmer. Die Waende und der Boden des Zimmers sind auf festen 
c  Temperaturen gehalten, in den Waenden befinden sich Fenster mit 
c  eigenen Temperaturwerten. Als Waermequelle dient ein Heizkoerper in 
c  der linken unteren Zimmerecke.  
c  Die Geometrie des Zimmers, die Staerke der Waermequelle und die
c  physikalischen Eigenschaften von Luft sind in Parameterstetements
c  festgehalten. Die Waermequelle wurde so parametriesiert, dass
c  die waehrend eines Tages abgegebene Waermemenge die Temperatur
C  des Zimmers um 10 Grad erhoehen wuerde, wenn das Zimmer vollstaedig
c  waermeisoliert waere.
c  Die Diskretisierung erfolgt mit dem 5-Punkte Stern.
c
c=====7==1=========2=========3=========4=========5=========6=========7==
      implicit none
      integer             col(*), rpt(*)
      integer             n, nmat,  nzero, istat
      double precision    val(*), rhs(*)
      double precision    h
      integer             nmax, nmat_max, nzero_max
      parameter         ( nmax = 950 )
      double precision    q(nmax*nmax), ru(0:nmax), 
     :                    ro(0:nmax), rw(0:nmax)
      integer             nm, nm2, i, j, k, l, ival
      double precision    vol_zimmer, vol_heizung, wq
      double precision    zimmer_mass, heizung_tiefe, heizung_hoehe,
     :                    fenster_unten, fenster_oben, 
     :                    temp_boden, temp_wand, temp_fenster,
     :                    quelle, kappa
      parameter         (
     :                    zimmer_mass    =  3.0,
     :                    heizung_tiefe  =  0.5,
     :                    heizung_hoehe  =  0.8,
     :                    fenster_unten  =  1.0,
     :                    fenster_oben   =  2.0,
     :                    temp_boden     = 10.0,
     :                    temp_wand      = 20.0,
     :                    temp_fenster   =  5.0,
     :                    quelle         =  50.0 / (24. * 3600),  
     :                    kappa          =  25.9 * 1.e-6
     :                                                            )

      nmat_max = nmat
      nzero_max = nzero
      if (n.gt.nmax) then
         write (6,*) 'n (',n,') groesser als nmax (',nmax,')'
         istat = -3
         return
      end if
      
      if (n.le.2) then
         write (6,*) 'n (',n,') zu klein' 
         istat = -3
         return
      end if

      h =  zimmer_mass / real(n)
      nm = n-1
      nm2 = nm**2
      nmat = nm2
      nzero = 5 * nmat
      if (nmat.gt.nmat_max.or.nzero.gt.nzero_max) then
         write(6,*)'nmat or nonzero larger then declared'
         istat = -3
         return
      end if
c--------- Waermequelle --------------------
      vol_zimmer = zimmer_mass**3
      vol_heizung = heizung_tiefe * heizung_hoehe * zimmer_mass
      wq = quelle * vol_zimmer / vol_heizung * 1. / kappa * h**2
      do i = 1 , n-1
         do j = 1 , n-1
            q(i+(j-1)*nm) = 0.
            if (i*h.le.heizung_tiefe.and.
     :          j*h.le.heizung_hoehe)      q(i+(j-1)*nm) = wq
         end do
      end do
      
c--------- Randbedingungen -----------
      do i =  0 , n
         ru(i) = temp_boden
         ro(i) = temp_wand
      end do
      
      do j = 1 , nm
         rw(j) = temp_wand
         if (j*h.ge.fenster_unten.and.j*h.le.fenster_oben)
     :      rw(j) = temp_fenster
      end do
c      write(6,1000)(ru(i),i=0,n)
c      write(6,1000)(ro(i),i=0,n)
c      write(6,1000)(rw(i),i=0,n)
c 1000 format(21f4.0)
c---------- rechte Seite ---------------
      do k = 1 , nm2
         rhs(k) = q(k)
      end do

c--------- Koeffizientenmatrix
      ival = 1
      do j = 1 , nm
         do i = 1 , nm
            k = i + (j-1)*nm
            rpt(k) = ival
            if (j-1.eq.0) then
               rhs(k) = rhs(k) + ru(i)
            else
               val(ival)= -1.
               col(ival) = k-nm
               ival = ival + 1
            end if
            if (i-1.eq.0) then
                rhs(k) = rhs(k) + rw(j)
            else 
               val(ival)= -1.
               col(ival) = k-1
               ival = ival + 1
            end if
            val(ival)= 4.
            col(ival) = k
            ival = ival + 1
           if (i+1.eq.n) then
                rhs(k) = rhs(k) + rw(j)
            else
               val(ival)= -1.
               col(ival) = k+1
               ival = ival + 1
            end if
            if (j+1.eq.n) then
                rhs(k) = rhs(k) + ro(i)
            else 
               val(ival)= -1.
               col(ival) = k+nm
               ival = ival + 1
            end if
         end do
      end do
      nmat = k
      rpt(nmat+1) = ival
      nzero = ival - 1
      return
      end


