      subroutine matvec(v,u,ipar)
      implicit none
      real                u(*), v(*)
      integer             ipar(*)
      integer             nloc, i, l
      INCLUDE "include_parameters"
c------------- data structure for distributed matrix in 
c------------- global address space 
      integer             col(matlen), rpt(veclen+1)
      real                val(matlen)
      common            / matrix / val, col, rpt
      real             w(veclen+extlen)
      common         / copy / w

      call pack_extern_vars(v,w,ipar)

      nloc = ipar(4)
      
      do i = 1 , nloc
         u(i) = 0.0d0
         do l = rpt(i) , rpt(i+1)-1
            u(i) = u(i) + val(l) * w(col(l))
         end do
      end do

      return
      end

      subroutine pack_extern_vars(v,x,ipar)
      real             v(*), x(*)
      integer          ipar(*)
      INCLUDE "include_parameters"
c------------- data structure for distributed matrix in 
c------------- global address space 
      integer             col(matlen), rpt(veclen+1)
      real                val(matlen)
      common            / matrix / val, col, rpt
c------------- data structure for externel variables in 
c------------- global address space 
      integer             get_list(extlen), get_id(proclen), 
     :                    get_pt(proclen), get_nb(proclen), nget
      integer             put_list(extlen), put_id(proclen),
     :                    put_pt(proclen), put_nb(proclen), nput
      common   / extern / get_list, get_id, get_pt, get_nb, nget, 
     :                    put_list, put_id, put_pt, put_nb, nput
      real             data(veclen)
      common         / packbuff / data
      integer          mypid, nloc, nproc
      integer          sp, rp, lp, n, ia ,ie, i 

      nloc = ipar(4)
      nproc = ipar(6)
      mypid = ipar(7)
      do i = 1 , nloc
         x(i) = v(i)
      end do
      do ig = 1 , nput
         call barrier()
         sp = put_id(ig)
         if (sp.ne.-1) then
            ia = put_pt(ig)
            n = put_nb(ig)
            do i = 1 , n
               data(i) = x(put_list(ia-1+i))
            end do 
         end if
         call barrier()
         rp = get_id(ig)
         if (rp.ne.-1) then
            ia = get_pt(ig)
            n = get_nb(ig)
            call shmem_get(x(nloc+ia),data,n,rp)
         end if
      end do

      return
      end

      real function psnrm(n,x,ipar)
c=====7==1=========2=========3=========4=========5=========6=========7==
c
c                               psnorm 
c
c     performs the parallel vector norm
c                   
c=====7==1=========2=========3=========4=========5=========6=========7==
      implicit none

c---------- dummy variables
      integer        n, ipar(*)
      real           x(*)
c---------- external 
      external pssum
c---------- local variables
      integer         i
      real           s, sdot
c      external ddot
      intrinsic sqrt
c============ end of declaration part ================================


      s = sdot(n,x,1,x,1)
      call pssum(1,s,ipar)
      psnrm =  sqrt(s)
      return
      end

      subroutine pssum(n,data,ipar)
      implicit none
      INCLUDE "mpp/shmem.fh"
      real             data(*)
      integer          n, ipar(*)
      integer buflen
      parameter (buflen=1000)
      real             res(buflen)
      real             pwrk(buflen/2+1+SHMEM_REDUCE_MIN_WRKDATA_SIZE)
      common /sumcom/res,pwrk
      integer psync(SHMEM_COLLECT_SYNC_SIZE)
      data psync /SHMEM_COLLECT_SYNC_SIZE*SHMEM_SYNC_VALUE/
      integer          nproc,i,ierr,num_pes

      if (n.gt.buflen) then
         write(6,*)'buffer too small for globsum:'
         write(6,*)'buflen =',buflen,' datasize =',n
         stop
      end if

      nproc = ipar(6)
      do i = 1 , n
         res(i) = data(i)
      end do
      call shmem_real8_sum_to_all(res,res,n,0,0,nproc,pwrk,psync)
      do i = 1 , n
         data(i) = res(i)
      end do

      return
      end

      subroutine preconl
      return
      end

      subroutine preconr
      return
      end

      subroutine progress
      return
      end

      subroutine scale_matrix(ipar)
      implicit none
c=====7==1=========2=========3=========4=========5=========6=========7==
      integer             ipar(*)

      INCLUDE "include_parameters"
c------------- data structure for distributed matrix in 
c------------- global address space 
      integer             col(matlen), rpt(veclen+1)
      real                val(matlen)
      common            / matrix / val, col, rpt
c------------- data structure for scale
      real                s(veclen)
      common            / scale / s

      integer          nloc, i, l
      real             w(veclen+extlen)
      common         / copy / w

      call pack_extern_vars(s,w,ipar)

      nloc = ipar(4)

      do i = 1 , nloc
         do l = rpt(i) , rpt(i+1)-1
            val(l) = w(i) * val(l) * w(col(l))
         end do
      end do

      return
      end

      subroutine scale_vector(x,ipar)
      implicit none
c=====7==1=========2=========3=========4=========5=========6=========7==
      integer          ipar(*)
      real             x(*)

      INCLUDE "include_parameters"
c------------- data structure for scale
      real                s(veclen)
      common            / scale / s
      integer nloc, i

      nloc = ipar(4)
      do i = 1 , nloc
         x(i) = x(i) * s(i)
      end do

      return
      end
