      SUBROUTINE PIMSCGEV(X,B,WRK,IPAR,SPAR,MATVEC,PRECONL,PRECONR,
     +                    PSSUM,PSNRM,PROGRESS)
      IMPLICIT NONE

*           PIM -- The Parallel Iterative Methods package
*           ---------------------------------------------
*
*                      Rudnei Dias da Cunha
*     National Supercomputing Centre and Mathematics Institute
*         Universidade Federal do Rio Grande do Sul, Brasil
*
*                          Tim Hopkins
*     Computing Laboratory, University of Kent at Canterbury, U.K.
*
* ----------------------------------------------------------------------
*
*     .. Parameters ..
      REAL ZERO
      PARAMETER (ZERO=0.0E0)
      REAL ONE
      PARAMETER (ONE=1.0E0)
      REAL TWO
      PARAMETER (TWO=2.0E0)
      INTEGER IPARSIZ
      PARAMETER (IPARSIZ=13)
      INTEGER SPARSIZ
      PARAMETER (SPARSIZ=6)
*     ..
*     .. Array Arguments ..
      REAL B(*),SPAR(SPARSIZ),WRK(*),X(*)
      INTEGER IPAR(IPARSIZ)
*     ..
*     .. Function Arguments ..
      REAL PSNRM
      EXTERNAL PSNRM
*     ..
*     .. Subroutine Arguments ..
      EXTERNAL MATVEC,PRECONL,PRECONR,PROGRESS,PSSUM
*     ..
*     .. Local Scalars ..
      REAL A1,ALPHA,B1,BETA,C,D,D1,DELTA,E,EPSILON,EXITNORM,F,F0,F1,G,H,
     +     H0,H1,MACHEPS,MU1,MUN,RDOTR,RDOTR0,RHO,RHO0,RHSSTOP,TA,TA1,
     +     TB,TB0,XI,XI0
      INTEGER BASISDIM,BLKSZ,CNVRTX,IDT,IET,IP,IR,IRSLT,IS,ITNO,IW,
     +        IXOLD,IZ,LDA,LOCLEN,MAXIT,N,NPROCS,PRECONTYPE,PROCID,
     +        STATUS,STEPERR,STOPTYPE
*     ..
*     .. Local Arrays ..
      REAL DOTS(2)
*     ..
*     .. External Functions ..
      REAL BISECTION,SDOT,SSETRHSSTOP
      EXTERNAL BISECTION,SDOT,SSETRHSSTOP
*     ..
*     .. External Subroutines ..
      EXTERNAL PIMSGETPAR,SAXPY,SCOPY,SMACHCONS,STOPCRIT
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC ABS,MAX,MIN,SQRT
*     ..

*     .. Scalars in Common ..
      REAL OLIMIT,OVERFLOW,ULIMIT,UNDERFLOW
*     ..
*     .. Common blocks ..
      COMMON /MACHCONS1/UNDERFLOW,OVERFLOW
      COMMON /MACHCONS2/ULIMIT,OLIMIT
*     ..
      CALL SMACHCONS('M',MACHEPS)
      CALL SMACHCONS('U',UNDERFLOW)
      ULIMIT = SQRT(UNDERFLOW)*TWO
      CALL SMACHCONS('O',OVERFLOW)
      OLIMIT = SQRT(OVERFLOW)/TWO

      CALL PIMSGETPAR(IPAR,SPAR,LDA,N,BLKSZ,LOCLEN,BASISDIM,NPROCS,
     +                PROCID,PRECONTYPE,STOPTYPE,MAXIT,ITNO,STATUS,
     +                STEPERR,EPSILON,EXITNORM)

*  Check consistency of preconditioning and stop types
      IF (((PRECONTYPE.EQ.0).OR. (PRECONTYPE.EQ.2)) .AND.
     +    (STOPTYPE.EQ.6)) THEN
          ITNO = 0
          STATUS = -4
          STEPERR = 0
          GO TO 9999

      END IF

*  Does not need conversion Y=Q2X for residual
      CNVRTX = 0

*  Set indices for mapping local vectors into wrk
      IR = 1
      IP = IR + LOCLEN
      IW = IP + LOCLEN
      IZ = IW + LOCLEN
      IS = IZ + LOCLEN
      IXOLD = IS + LOCLEN
      IDT = IXOLD + LOCLEN
      IET = IDT + MAXIT + 1

*  Set rhs of stopping criteria
      RHSSTOP = SSETRHSSTOP(B,WRK(IR),EPSILON,IPAR,PRECONL,PSNRM)

*  1. r=Q1(b-AQ2x)
      IF (STOPTYPE.NE.6) THEN
          IF (PRECONTYPE.EQ.0) THEN
*     r=b-Ax
              CALL SCOPY(LOCLEN,B,1,WRK(IR),1)
              CALL MATVEC(X,WRK(IW),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IW),1,WRK(IR),1)

          ELSE IF (PRECONTYPE.EQ.1) THEN
*     r=Q1(b-Ax)
              CALL SCOPY(LOCLEN,B,1,WRK(IZ),1)
              CALL MATVEC(X,WRK(IW),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IW),1,WRK(IZ),1)
              CALL PRECONL(WRK(IZ),WRK(IR),IPAR)

          ELSE IF (PRECONTYPE.EQ.2) THEN
*     r=b-AQ2x
              CALL SCOPY(LOCLEN,B,1,WRK(IR),1)
              CALL PRECONR(X,WRK(IW),IPAR)
              CALL MATVEC(WRK(IW),WRK(IZ),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IZ),1,WRK(IR),1)

          ELSE IF (PRECONTYPE.EQ.3) THEN
*     r=Q1(b-AQ2x)
              CALL SCOPY(LOCLEN,B,1,WRK(IP),1)
              CALL PRECONR(X,WRK(IW),IPAR)
              CALL MATVEC(WRK(IW),WRK(IZ),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IZ),1,WRK(IP),1)
              CALL PRECONL(WRK(IP),WRK(IR),IPAR)
          END IF

      ELSE
*     r has been set to Qb in the call to dsetrhsstop
          IF (PRECONTYPE.EQ.1) THEN
*     r=Q1(b-Ax)
              CALL MATVEC(X,WRK(IW),IPAR)
              CALL PRECONL(WRK(IW),WRK(IZ),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IZ),1,WRK(IR),1)

          ELSE IF (PRECONTYPE.EQ.3) THEN
*     r=Q1(b-AQ2x)
              CALL PRECONR(X,WRK(IZ),IPAR)
              CALL MATVEC(WRK(IZ),WRK(IW),IPAR)
              CALL PRECONL(WRK(IW),WRK(IZ),IPAR)
              CALL SAXPY(LOCLEN,-ONE,WRK(IZ),1,WRK(IR),1)
          END IF

      END IF

*  2. p=r
      CALL SCOPY(LOCLEN,WRK(IR),1,WRK(IP),1)

*  3. rdotr=dot(r,r)
      DOTS(1) = SDOT(LOCLEN,WRK(IR),1,WRK(IR),1)
      CALL PSSUM(1,DOTS,IPAR)
      RDOTR = DOTS(1)

*  4. w=Q1AQ2p
      IF (PRECONTYPE.EQ.0) THEN
          CALL MATVEC(WRK(IP),WRK(IW),IPAR)

      ELSE IF (PRECONTYPE.EQ.1) THEN
          CALL MATVEC(WRK(IP),WRK(IZ),IPAR)
          CALL PRECONL(WRK(IZ),WRK(IW),IPAR)

      ELSE IF (PRECONTYPE.EQ.2) THEN
          CALL PRECONR(WRK(IP),WRK(IZ),IPAR)
          CALL MATVEC(WRK(IZ),WRK(IW),IPAR)

      ELSE IF (PRECONTYPE.EQ.3) THEN
          CALL PRECONR(WRK(IP),WRK(IW),IPAR)
          CALL MATVEC(WRK(IW),WRK(IZ),IPAR)
          CALL PRECONL(WRK(IZ),WRK(IW),IPAR)
      END IF

* 5. xi=dot(p,w)
      DOTS(1) = SDOT(LOCLEN,WRK(IP),1,WRK(IW),1)
      CALL PSSUM(1,DOTS,IPAR)
      XI = DOTS(1)

* 6. Initialise data for eigenvalue estimation
      MU1 = ZERO
      MUN = ZERO
      RHO = SQRT(RDOTR)
*  Use rdotr==rho**2
      IF (RDOTR.EQ.ZERO) THEN
	  ITNO = 0
          STATUS = -3
          STEPERR = 6
          GO TO 9999

      END IF

      TA1 = XI/RDOTR
      WRK(IDT) = TA1
      TB = ZERO

*  Loop
      STATUS = 0
      EXITNORM = -ONE
      STEPERR = -1
      DO 20 ITNO = 1,MAXIT

*  7. alpha=rdotr/xi
          IF (XI.EQ.ZERO) THEN
              STATUS = -3
              STEPERR = 7
              GO TO 9999

          END IF

          ALPHA = RDOTR/XI

*  8. x=x+alpha*p
          CALL SCOPY(LOCLEN,X,1,WRK(IXOLD),1)
          CALL SAXPY(LOCLEN,ALPHA,WRK(IP),1,X,1)

*  9. r=r-alpha*w
          CALL SAXPY(LOCLEN,-ALPHA,WRK(IW),1,WRK(IR),1)

* 10. check stopping criterion
          CALL STOPCRIT(B,WRK(IR),WRK(IZ),X,WRK(IXOLD),WRK(IS),RHSSTOP,
     +                  CNVRTX,EXITNORM,STATUS,IPAR,MATVEC,MATVEC,
     +                  PRECONR,PSSUM,PSNRM)

*  Call monitoring routine
          CALL PROGRESS(LOCLEN,ITNO,EXITNORM,X,WRK(IR),WRK(IZ))

          IF (STATUS.EQ.-5) THEN
              STEPERR = 10
              GO TO 9999
          ELSE IF (STATUS.EQ.0) THEN
              GO TO 9999
          END IF
* 11. s=Q1AQ2r
          IF (PRECONTYPE.EQ.0) THEN
              CALL MATVEC(WRK(IR),WRK(IS),IPAR)

          ELSE IF (PRECONTYPE.EQ.1) THEN
              CALL MATVEC(WRK(IR),WRK(IZ),IPAR)
              CALL PRECONL(WRK(IZ),WRK(IS),IPAR)

          ELSE IF (PRECONTYPE.EQ.2) THEN
              CALL PRECONR(WRK(IR),WRK(IZ),IPAR)
              CALL MATVEC(WRK(IZ),WRK(IS),IPAR)

          ELSE IF (PRECONTYPE.EQ.3) THEN
              CALL PRECONR(WRK(IR),WRK(IS),IPAR)
              CALL MATVEC(WRK(IS),WRK(IZ),IPAR)
              CALL PRECONL(WRK(IZ),WRK(IS),IPAR)
          END IF

* 12. rdotr=dot(r,r)
          RDOTR0 = RDOTR
          DOTS(1) = SDOT(LOCLEN,WRK(IR),1,WRK(IR),1)

* 13. delta=dot(r,s)
          DOTS(2) = SDOT(LOCLEN,WRK(IR),1,WRK(IS),1)

*  Accumulate simultaneously partial values
          CALL PSSUM(2,DOTS,IPAR)
          RDOTR = DOTS(1)
          DELTA = DOTS(2)

* 14. beta=rdotr/rdotr0
          IF (RDOTR0.EQ.ZERO) THEN
              STATUS = -3
              STEPERR = 14
              GO TO 9999

          END IF

          BETA = RDOTR/RDOTR0

* 15. p=r+beta*p
          CALL SCOPY(LOCLEN,WRK(IP),1,WRK(IZ),1)
          CALL SCOPY(LOCLEN,WRK(IR),1,WRK(IP),1)
          CALL SAXPY(LOCLEN,BETA,WRK(IZ),1,WRK(IP),1)

* 16. w=s+beta*w
          CALL SCOPY(LOCLEN,WRK(IW),1,WRK(IZ),1)
          CALL SCOPY(LOCLEN,WRK(IS),1,WRK(IW),1)
          CALL SAXPY(LOCLEN,BETA,WRK(IZ),1,WRK(IW),1)

* 17. xi=delta-beta^2*xi
          XI0 = XI
          XI = DELTA - BETA**2*XI

* 18. compute estimates of eigenvalues
          TA = TA1
*  Use rdotr==rho**2.0
          RHO0 = RHO
          RHO = SQRT(RDOTR)

          TA1 = (BETA**TWO*XI0+XI)/RDOTR
          TB0 = TB
          TB = -XI0*BETA/ (RHO0*RHO)
          WRK(IDT+ITNO) = TA

*     Store square of subdiagonal elements to save operations in
*     the bisection routine
          WRK(IET+ITNO-1) = TB*TB

          IF (ITNO.EQ.1) THEN
*     Compute roots of the funression for the determinant
*       | a1 b1 |
*       | b1 a2 |
*     i.e., a quadratic equation
*     p(mu)=mu^2-(a1+a2)mu+a1*a2-b1^2 via quadratic formula; take
*     care of possible cancellation of digits
              A1 = TA + TA1
              B1 = TA*TA1 - WRK(IET+ITNO-1)
              IF (A1.LT.ZERO) THEN
                  D1 = SQRT(A1*A1-4.0E0*B1)
                  E = (A1+D1)/TWO
                  F = TWO*B1/ (A1+D1)
                  MU1 = MIN(E,F)
                  MUN = MAX(E,F)

              ELSE IF (A1.GT.ZERO) THEN
                  D1 = SQRT(A1*A1-4.0E0*B1)
                  E = TWO*B1/ (A1-D1)
                  F = (A1-D1)/TWO
                  MU1 = MIN(E,F)
                  MUN = MAX(E,F)

              ELSE
                  D1 = SQRT(-B1)
                  MU1 = -D1
                  MUN = D1
              END IF

*     Set [c,d] to [mu1,mun]
              C = MU1
              D = MUN

          ELSE

*     Update [c,d] via Gerschgorin's theorem
              G = ABS(TB)
              E = ABS(TB0) + G
              F = ABS(TA)
              F0 = F - E
              F1 = F + E
              H = ABS(TA1)
              H0 = H - G
              H1 = H + G
              C = MIN(C,MIN(F0,H0))
              D = MAX(D,MAX(F1,H1))

*     Update [mu1,mun] via bisection
              MU1 = BISECTION(MACHEPS,WRK(IDT),WRK(IET),ITNO+1,C,MU1,
     +              IRSLT)
              IF (IRSLT.EQ.-1) THEN
                  STATUS = -8
                  STEPERR = 18
                  GO TO 9999
              END IF
              IF (IRSLT.EQ.-2) THEN
                  STATUS = -9
                  STEPERR = 18
                  GO TO 9999
              END IF

              MUN = BISECTION(MACHEPS,WRK(IDT),WRK(IET),ITNO+1,MUN,D,
     +              IRSLT)
              IF (IRSLT.EQ.-1) THEN
                  STATUS = -10
                  STEPERR = 18
                  GO TO 9999
              END IF
              IF (IRSLT.EQ.-2) THEN
                  STATUS = -11
                  STEPERR = 18
                  GO TO 9999
              END IF


          END IF

   20 CONTINUE

      IF (ITNO.GT.MAXIT) THEN
          STATUS = -1
          ITNO = MAXIT
      END IF

 9999 CONTINUE

      IF ((PRECONTYPE.EQ.2) .OR. (PRECONTYPE.EQ.3)) THEN
          CALL SCOPY(LOCLEN,X,1,WRK(IZ),1)
          CALL PRECONR(WRK(IZ),X,IPAR)
      END IF

*  Set output parameters
      IPAR(11) = ITNO
      IPAR(12) = STATUS
      IPAR(13) = STEPERR
      SPAR(2) = EXITNORM
      SPAR(3) = MU1
      SPAR(4) = MUN
      SPAR(5) = 0.0E0
      SPAR(6) = 0.0E0

      RETURN

      END

      REAL FUNCTION BISECTION(SMACHEPS,A,B,K,Y,Z,IRSLT)
      IMPLICIT NONE

*  Computes the root of a function (the determinantal function
*  associated with the Lanczos's tridiagonal matrix) using the
*  bisection method.
*  Possible values of IRSLT: 0 - no floating-point errors
*                           -1 - underflow
*                           -2 - overflow
*     .. Parameters ..
      REAL ZERO
      PARAMETER (ZERO=0.0E0)
      REAL ONE
      PARAMETER (ONE=1.0E0)
      REAL TWO
      PARAMETER (TWO=2.0E0)
      INTEGER MAXIT
      PARAMETER (MAXIT=50)
*     ..
*     .. Scalar Arguments ..
      REAL SMACHEPS,Y,Z
      INTEGER IRSLT,K
*     ..
*     .. Array Arguments ..
      REAL A(*),B(*)
*     ..
*     .. Local Scalars ..
      REAL P2X,P2Y,X
      INTEGER IT
*     ..
*     .. External Functions ..
      REAL DETFUN
      EXTERNAL DETFUN
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC ABS
*     ..
*     .. Scalars in Common ..
      REAL OVERFLOW,UNDERFLOW
*     ..
*     .. Common blocks ..
      COMMON /MACHCONS1/UNDERFLOW,OVERFLOW
*     ..
      IRSLT = 0
      X = (Y+Z)/TWO
      IF (ABS(Y-Z).LE. (SMACHEPS* (ABS(Y)+ABS(Z)))) THEN
          BISECTION = X
          RETURN
      END IF

* Compute p2(y) once only
      P2Y = DETFUN(A,B,K,Y)
      IF (P2Y.EQ.UNDERFLOW) THEN
          BISECTION = X
          IRSLT = -1
          RETURN
      END IF
      IF (P2Y.EQ.OVERFLOW) THEN
          BISECTION = X
          IRSLT = -2
          RETURN
      END IF

      DO 10 IT = 1,MAXIT
          P2X = DETFUN(A,B,K,X)
          IF (P2X.EQ.UNDERFLOW) THEN
              BISECTION = X
              IRSLT = -1
              RETURN
          END IF
          IF (P2X.EQ.OVERFLOW) THEN
              BISECTION = X
              IRSLT = -2
              RETURN
          END IF

          IF (SIGN(ONE,P2X).NE.SIGN(ONE,P2Y)) THEN
* There is a root between x and y
* p2(y) remains the same for the next iteration
              Z = X

          ELSE IF (ABS(P2X).LT.SMACHEPS) THEN
* Close enought to zero, accept x as root
              BISECTION = X
              RETURN

          ELSE
* There is a root between x and z
* p2(y) becomes p2(x) for next iteration
              Y = X
              P2Y = P2X
          END IF

          IF (ABS(Y-Z).LE. (SMACHEPS* (ABS(Y)+ABS(Z)))) THEN
              BISECTION = X
              RETURN
          END IF
          X = (Y+Z)/TWO
   10 CONTINUE
      BISECTION = X
      RETURN

      END

      REAL FUNCTION DETFUN(A,B,K,X)
      IMPLICIT NONE


*  Compute determinant function at x; the squares of
*  the subdiagonal elements of T_{k} are stored in b
*     .. Scalar Arguments ..
      REAL X
      INTEGER K
*     ..
*     .. Array Arguments ..
      REAL A(*),B(*)
*     ..
*     .. Local Scalars ..
      REAL P0,P1,P2
      INTEGER I
*     ..
*     .. Scalars in Common ..
      REAL OLIMIT,OVERFLOW,ULIMIT,UNDERFLOW
*     ..
*     .. Common blocks ..
      COMMON /MACHCONS1/UNDERFLOW,OVERFLOW
      COMMON /MACHCONS2/ULIMIT,OLIMIT
*     ..
*     .. Parameters ..
      REAL ZERO
      PARAMETER (ZERO=0.0E0)
*     ..
*     .. Intrinsic Functions ..
      INTRINSIC ABS
*     ..
      P1 = A(1) - X
      P2 = (A(2)-X)*P1 - B(1)
      DO 10 I = 3,K
          P0 = P1
          P1 = P2
          P2 = (A(I)-X)*P1 - B(I-1)*P0
          IF ((ABS(P2).GT.ZERO) .AND. (ABS(P2).LE.ULIMIT)) THEN
              DETFUN = UNDERFLOW
              RETURN
          END IF
          IF (ABS(P2).GE.OLIMIT) THEN
              DETFUN = OVERFLOW
              RETURN
          END IF

   10 CONTINUE

      DETFUN = P2
      RETURN

      END
